package org.expeditee.items;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.Toolkit;
import java.awt.geom.Point2D;
import java.awt.image.CropImageFilter;
import java.awt.image.FilteredImageSource;
import java.awt.image.ImageObserver;
import java.io.File;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLDecoder;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;

import org.expeditee.gui.FrameGraphics;
import org.expeditee.gui.FrameMouseActions;
import org.expeditee.settings.UserSettings;

/**
 * This class represents an Image loaded from a file which is shown on the
 * screen. Loading of the Image from disk occurs in the constructor, and takes
 * approximately one second per mb of the Image file size. <br>
 * <br>
 * Currently Supported (Tested) Image formats:<br>
 * BMP<br>
 * JPG<br>
 * GIF<br>
 * GIF (Animated)<br>
 * <br>
 * Currently only the default size of the Image is supported, but future
 * versions may support scaling.
 * 
 * @author jdm18
 * 
 */
public class Picture extends Item {
	
	protected static final Image imageNotFound;
	protected static final Polygon imageNotFoundPolygon;
	static {
		Image iNF = null;
		try {
			URL image = ClassLoader.getSystemResource("org/expeditee/resources/images/imageNotFound.png");
			if(image != null) {
				iNF = Toolkit.getDefaultToolkit().getImage(image);
			} else {
				System.err.println("Failed to load default image");
			}
		} catch(Exception e) {
			e.printStackTrace();
		}
		imageNotFound = iNF;
		imageNotFoundPolygon = new Polygon();
		imageNotFoundPolygon.addPoint(-1,  -1);
		imageNotFoundPolygon.addPoint(65, -1);
		imageNotFoundPolygon.addPoint(65,  65);
		imageNotFoundPolygon.addPoint(-1, 65);
		imageNotFoundPolygon.addPoint(-1,  65);
		System.out.println("Loaded static members");
	}

	private static final int MINIMUM_WIDTH = 10;
	public static final int WIDTH = 0;
	public static final int RATIO = 1;

	protected Image _image = null;
	protected Image _croppedImage = null;

	private int _scaleType = RATIO;

	private float _scale = 1.0f;

	// Points relative to this Picture's cropping, for creating further cropped Pictures
	private Point _cropStart = null;
	private Point _cropEnd = null;

	// Crop data of this Picture
	private Point _start = new Point(0, 0);
	private Point _end = new Point(0, 0);

	private boolean _showCropping = false;

	private String _path = "";

	// used to repaint animated GIF images, among other things.
	protected ImageObserver _imageObserver = null;

	/**
	 * Creates a new Picture from the given path. The ImageObserver is optional
	 * and can be set to NULL. <br>
	 * Note: It is assumed that the file described in path has already been
	 * checked to exist.
	 * 
	 * @param path
	 *            The Path of the Image to load from disk.
	 * @param observer
	 *            The ImageObserver to assign when painting the Image on the
	 *            screen.
	 */
	
	public Picture(int id) {
		super();
		setID(id);
	}
	
	public Picture(String path, ImageObserver observer, int id) {
		super();
		if(id == -1) {
			id = this.getParentOrCurrentFrame().getNextItemID();
		}
		setID(id);
		_imageObserver = observer;
		this.setImageSource(path);
	}
	
	public Picture(String path, ImageObserver observer) {
		this(path, observer, -1);
	}
	
	public Picture(String path, Text source, String size, ImageObserver observer) {
		this(path, observer, source.getID());
		Item.DuplicateItem(source, this);
		this.parseSize(size);
		updatePolygon();
	}
	
	/**
	 * Parse string input given in an @i tag
	 */
	protected void parseSize(String size) {
		// don't check this anymore since it's only called from constructors now
		// if (_end.x != 0 || _end.y != 0)
		//	return;

		// set the default values for start and end
		_start.setLocation(0, 0);
		if (_image == null)
			_end.setLocation(0, 0);
		else
			_end.setLocation(_image.getWidth(null), _image.getHeight(null));
		size = size.trim();
		String[] values = size.split("\\s+");
		// Now get the cropping values if there are any
		try {
			if (values.length > 2) {
				int startX = Integer.parseInt(values[1]);
				int startY = Integer.parseInt(values[2]);
				_start = new Point(startX, startY);
				if (values.length > 4) {
					int endX = Integer.parseInt(values[3]);
					int endY = Integer.parseInt(values[4]);
					_end = new Point(endX, endY);
				}
			}
		} catch (Exception e) {
		}

		try {
			if (size.length() == 0) {
				size = "" + _image.getWidth(null);
				return;
			}
			size = values[0];
			// parse width or ratio from text
			if (size.contains(".")) {
				// this is a ratio
				_scale = Float.parseFloat(size);
				_scaleType = RATIO;
			} else if (size.length() > 0) {
				// this is an absolute width
				int width = Integer.parseInt(size);
				_scaleType = WIDTH;
				setWidth(width);
			}
		} catch (Exception e) {
			_scale = 1F;
		}
	}


	/**
	 * Set the start point for generating a new cropped Picture from this Picture
	 */
	public void setStartCrop(int x, int y) {
		invalidateCroppedArea();
		_cropStart = new Point(x - getX(), y - getY());
		invalidateCroppedArea();
	}

	/**
	 * Set the end point for generating a new cropped Picture from this Picture
	 */
	public void setEndCrop(int x, int y) {
		invalidateCroppedArea();
		_cropEnd = new Point(x - getX(), y - getY());
		invalidateCroppedArea();
	}

	private void invalidateCroppedArea() {
		if (_cropStart != null && _cropEnd != null) {
			Point topLeft = getTopLeftCrop();
			Point bottomRight = getBottomRightCrop();
			int startX = getX() + topLeft.x - _highlightThickness;
			int startY = getY() + topLeft.y - _highlightThickness;
			int border = 2 * _highlightThickness;
			invalidate(new Rectangle(startX, startY, bottomRight.x - topLeft.x
					+ 2 * border, bottomRight.y - topLeft.y + 2 * border));
			invalidateAll();
		} else {
			invalidateAll();
		}
	}

	/**
	 * Get the top left point (_start) for generating a new cropped Picture from this Picture
	 */
	public Point getTopLeftCrop() {
		return new Point(Math.min(_cropStart.x, _cropEnd.x), Math.min(
				_cropStart.y, _cropEnd.y));
	}

	/**
	 * Get the bottom right point (_end) for generating a new cropped Picture from this Picture
	 */
	public Point getBottomRightCrop() {
		return new Point(Math.max(_cropStart.x, _cropEnd.x), Math.max(
				_cropStart.y, _cropEnd.y));
	}

	public void setShowCrop(boolean value) {
		// invalidateCroppedArea();
		_showCropping = value;
		invalidateCroppedArea();
	}

	public boolean isCropTooSmall() {
		if (_cropStart == null || _cropEnd == null || _image == imageNotFound)
			return true;

		int cropWidth = Math.abs(_cropEnd.x - _cropStart.x);
		int cropHeight = Math.abs(_cropEnd.y - _cropStart.y);

		return cropWidth < MINIMUM_WIDTH || cropHeight < MINIMUM_WIDTH;
	}

	public void clearCropping() {
		invalidateCroppedArea();
		_cropStart = null;
		_cropEnd = null;
		setShowCrop(false);
	}
	
	public int getStartX() {
		return _image == imageNotFound ? 0 : _start.x;
	}
	
	public int getStartY() {
		return _image == imageNotFound ? 0 : _start.y;
	}
	
	public int getEndX() {
		return _image == imageNotFound ? 64 : _end.x == 0 ? _image.getWidth(null) : _end.x;
	}
	
	public int getEndY() {
		return _image == imageNotFound ? 64 : _end.y == 0 ? _image.getHeight(null) : _end.y;
	}

	public void updatePolygon() {
		if (_image == null) {
			refresh();
		}
		if (_image == imageNotFound) {
			_poly = new Polygon(imageNotFoundPolygon.xpoints, imageNotFoundPolygon.ypoints, imageNotFoundPolygon.npoints);
			_poly.translate(getX(), getY());
			// _start = new Point(0, 0);
			// _end = new Point(_image.getWidth(null), _image.getHeight(null));
			// _scale = 1f;
			return;
		}
		
		_poly = new Polygon();

		if (_cropStart == null || _cropEnd == null) {
			int width = getWidth();
			int height = getHeight();

			int xdiff = -MARGIN_RIGHT; // -getLeftMargin();

			// extra pixel around the image so the highlighting is visible
			_poly.addPoint(getX() + 1 + xdiff, getY() - 1);
			_poly.addPoint(getX() + width, getY() - 1);
			_poly.addPoint(getX() + width, getY() + height);
			_poly.addPoint(getX() + 1 + xdiff, getY() + height);
		} else {
			Point topLeft = getTopLeftCrop();
			Point bottomRight = getBottomRightCrop();
			Rectangle clip = new Rectangle(topLeft.x + getX(),
					topLeft.y + getY(), bottomRight.x - topLeft.x,
					bottomRight.y - topLeft.y).getBounds();
			_poly.addPoint((int) clip.getMinX() - 1, (int) clip.getMinY() - 1);
			_poly.addPoint((int) clip.getMinX() - 1, (int) clip.getMaxY());
			_poly.addPoint((int) clip.getMaxX(), (int) clip.getMaxY());
			_poly.addPoint((int) clip.getMaxX(), (int) clip.getMinY() - 1);

		}
	}

	@Override
	public double getEnclosedArea() {
		return getWidth() * getHeight();
	}

	@Override
	public void setWidth(Integer width) {
		_scaleType = WIDTH;
		_scale = width * 1F / (getUnscaledWidth());
	}

	public Point getStart() {
		return _start;
	}
	
	public Point getEnd() {
		return _end;
	}
	
	/**
	 * Gets the width with which the picture is displayed on the screen.
	 */
	@Override
	public Integer getWidth() {
		return Math.round((getUnscaledWidth()) * _scale);
	}

	/**
	 * Gets the height with which the picture is displayed on the screen.
	 */
	@Override
	public int getHeight() {
		return Math.round((getUnscaledHeight()) * _scale);
	}

	/**
	 * Dont paint links in audience mode for images.
	 */
	@Override
	protected void paintLink(Graphics2D g) {
		if (FrameGraphics.isAudienceMode())
			return;
		super.paintLink(g);
	}

	@Override
	public void paint(Graphics2D g) {
		// System.out.println("drawing image");

		if(_image == null) {
			return;
		}
		
		int width = _image == imageNotFound ? 64 : getWidth();
		int height = _image == imageNotFound ? 64 : getHeight();
		int startX = getStartX();
		int startY = getStartY();
		int endX = getEndX();
		int endY = getEndY();
		
		// System.out.println(width + " : " + height);

		paintLink(g);

		int dX1 = getX();
		int dY1 = getY();
		int dX2 = getX() + width;
		int dY2 = getY() + height;

		// if we are showing the cropping, then show the original as transparent
		if (_showCropping && !isCropTooSmall()) {
			// show the full image as transparent
			float alpha = .5f;
			g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,
					alpha));
			g.drawImage(_image, dX1, dY1, dX2, dY2, startX, startY, endX, endY, _imageObserver);

			g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,
					1.0f));
			// show the cropped area normally
			Point topLeft = getTopLeftCrop();
			Point bottomRight = getBottomRightCrop();
			Shape clip = new Rectangle(dX1 + topLeft.x, dY1 + topLeft.y,
					bottomRight.x - topLeft.x, bottomRight.y - topLeft.y);
			g.setColor(getPaintHighlightColor());
			g.draw(clip);
			g.setClip(clip);

			g.drawImage(_image, dX1, dY1, dX2, dY2, startX, startY, endX, endY, _imageObserver);
			g.draw(clip);
			// if the image is cropped, but we are not showing the cropping
			// otherwise, paint normally
		} else {
			g.drawImage(_image, dX1, dY1, dX2, dY2, startX, startY, endX, endY, _imageObserver);
		}

		if (hasVisibleBorder()) {
			g.setColor(getPaintBorderColor());
			Stroke borderStroke = new BasicStroke(getThickness(), CAP, JOIN);
			g.setStroke(borderStroke);
			g.drawPolygon(getPolygon());
		}

		if (isHighlighted()) {
			Stroke borderStroke = new BasicStroke(1, CAP, JOIN);
			g.setStroke(borderStroke);
			g.setColor(getHighlightColor());
			g.drawPolygon(getPolygon());
		}
		
		//System.out.print("p_");	
	}

	@Override
	public Color getHighlightColor() {
		if (_highlightColor.equals(getBorderColor()))
			return ALTERNATE_HIGHLIGHT;
		return _highlightColor;
	}

	@Override
	public int setHighlightColor() {
		super.setHighlightColor();

		return Item.DEFAULT_CURSOR;
	}

	protected Picture createPicture() {
		return new Picture(_path, _imageObserver);
		// return ItemUtils.CreatePicture((Text) _source.copy(), _imageObserver);
	}

	@Override
	public Picture copy() {
		Picture p = createPicture();
		p._image = _image;
		p._mode = _mode;
		p.setPosition(getPosition());
		// Doing Duplicate item duplicates link mark which we dont want to do
		// when in audience mode because the linkMark will be copied incorrectly
		// Get all properties from the source

		if (!isCropTooSmall() && _cropStart != null && _cropEnd != null) {
			assert (_cropEnd != null);
			// make the start be the top left
			// make the end be the bottom right
			Point topLeft = getTopLeftCrop();
			Point bottomRight = getBottomRightCrop();
			int startX = Math.round(topLeft.x / _scale) + getStartX();
			int startY = Math.round(topLeft.y / _scale) + getStartY();
			int endX = Math.round(bottomRight.x / _scale + getStartX());
			int endY = Math.round(bottomRight.y / _scale + getStartY());
			int width = _image.getWidth(null);
			int height = _image.getHeight(null);
			// adjust our start and end if the user has dragged outside of the
			// shape
			if (endX > width) {
				endX = width;
			}
			if (endY > height) {
				endY = height;
			}
			if (startX < 0) {
				startX = 0;
			}
			if (startY < 0) {
				startY = 0;
			}
			p._start = new Point(startX, startY);
			p._end = new Point(endX, endY);
			p.setPosition(topLeft.x + getX(), topLeft.y + getY());
		} else {
			p._start = _start;
			p._end = _end;
		}
		p._scale = _scale;
		p._scaleType = _scaleType;
		p._path = _path;

		p.updatePolygon();

		return p;
	}

	public float getScale() {
		return _scale;
	}

	public void setScale(float scale) {
		_scaleType = RATIO;
		_scale = scale;
	}

	@Override
	public float getSize() {
		// to let scaling work
		return UserSettings.ItemTemplate.getSize();
	}

	@Override
	public void setSize(float size) {
		float diff = size - getSize();
		float oldScale = _scale;

		float multiplier = (1000F + diff * 40F) / 1000F;
		_scale = _scale * multiplier;

		// picture must still be at least XX pixels wide
		if (getWidth() < MINIMUM_WIDTH) {
			_scale = oldScale;
		} else {
			this.translate(new Point2D.Float(FrameMouseActions.MouseX,
					FrameMouseActions.MouseY), multiplier);
		}
		updatePolygon();
		// Make sure items that are resized display the border
		invalidateAll();
	}

	@Override
	public void setAnnotation(boolean val) {
	}

	/**
	 * Returns the Image that this Picture object is painting on the screen.
	 * This is used by Frame to repaint animated GIFs.
	 * 
	 * @return The Image that this Picture object represents.
	 */
	public Image getImage() {
		return _image;
	}

	public Image getCroppedImage() {
		if (_image == null)
			return null;
		if (!isCropped()) {
			return _image;
		}
		
		return Toolkit.getDefaultToolkit().createImage(
				new FilteredImageSource(_image.getSource(),
						new CropImageFilter(getStartX(), getStartY(),
								getUnscaledWidth(), getUnscaledHeight())));
	}

	public int getUnscaledWidth() {
		return getEndX() - getStartX();
	}

	public int getUnscaledHeight() {
		return getEndY() - getStartY();
	}

	/**
	 * @return true if this is a cropped image.
	 */
	public boolean isCropped() {
		return (_start.x > 0 || _start.y > 0 || (_end.x > 0 && _end.x < _image.getWidth(null)) || (_end.y > 0 && _end.y < _image.getHeight(null)));
		// return _end.x != 0 || _end.y != 0 || _end.x != _image.getWidth(null) || _end.y != _image.getHeight(null) || _start.y != 0 || _start.x != 0;
	}

	public boolean refresh() {
		// try images subdirectory
		File file = null;

		for (String dir : org.expeditee.settings.UserSettings.ImageDirs) {
			file = new File(dir + _path);
			if (file.exists() && !file.isDirectory())
				break;
		}

		if (file == null || !file.exists() || file.isDirectory())
			file = new File(_path);

		// try relative path
		if (!file.exists() || file.isDirectory()) {
			URL picture = new Object().getClass().getResource(_path);

			// decode to remove %20 in windows folder names
			if (picture != null) {
				try {
					_path = URLDecoder.decode(picture.getFile(), "UTF-8");
				} catch (UnsupportedEncodingException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}

		} else
			_path = file.getPath();

		// if the image isn't found by now, give up.
		file = new File(_path);
		if (!file.exists() || file.isDirectory()) {
			// If the image could not be read
			_image = imageNotFound;
			return false;
		}

		// ImageIcon is faster, but cannot handle some formats
		// (notably.bmp) hence, we try this first, then if it fails we try
		// ImageIO
		try {
			_image = new ImageIcon(_path).getImage();
		} catch (Exception e) {
		}

		// if ImageIcon failed to read the image
		if (_image == null || _image.getWidth(null) <= 0) {
			try {
				_image = ImageIO.read(new File(_path));
			} catch (Exception e) {
				// e.printStackTrace();
				// Logger.Log(e);
				
				// If the image could not be read
				_image = imageNotFound;
				return false;
			}
		}
		return true;
	}

	@Override
	protected int getLinkYOffset() {
		return getBoundsHeight() / 2;
	}

	@Override
	public void setLinkMark(boolean state) {
		// TODO use the more efficient invalidiate method
		// The commented code below is not quite working
		// if(!state)
		// invalidateCommonTrait(ItemAppearence.LinkChanged);
		super.setLinkMark(state);
		// if(state)
		// invalidateCommonTrait(ItemAppearence.LinkChanged);
		invalidateAll();
	}

	@Override
	public void setActionMark(boolean state) {
		// if (!state)
		// invalidateCommonTrait(ItemAppearence.LinkChanged);
		super.setActionMark(state);
		// if (state)
		// invalidateCommonTrait(ItemAppearence.LinkChanged);
		invalidateAll();
	}

	@Override
	public String getName() {
		return _path;
	}

	@Override
	public void translate(Point2D origin, double ratio) {
		_scale *= ratio;
		super.translate(origin, ratio);
	}

	@Override
	public Rectangle[] getDrawingArea() {

		Rectangle[] da = super.getDrawingArea();

		if (getLink() != null || hasAction()) {
			Rectangle[] da2 = new Rectangle[da.length + 1];
			System.arraycopy(da, 0, da2, 0, da.length);
			da2[da.length] = getLinkPoly().getBounds();
			da2[da.length].translate(getX() - LEFT_MARGIN, getY()
					+ getLinkYOffset());
			da2[da.length].width += 2;
			da2[da.length].height += 2;
			da = da2;
		}

		return da;

	}
	
	public Item merge(Item merger, int mouseX, int mouseY) {
		return merger;
	}

	@Override
	public void scale(Float scale, int originX, int originY) {
		setScale(getScale() * scale);
		super.scale(scale, originX, originY);
	}
	
	public Integer getWidthToSave() {
		if(_scaleType == WIDTH) {
			return this.getWidth();
		}
		return null;
	}
	
	public Float getScaleToSave() {
		if(_scaleType == RATIO) {
			return this.getScale();
		}
		return null;
	}
	
	public Point getCropStart() {
		if (isCropped()) {
			return _start;
		}
		return null;
	}
	
	public Point getCropEnd() {
		if (isCropped()) {
			return _end;
		}
		return null;
	}
	
	public void setCropStart(Point cropStart) {
		invalidateCroppedArea();
		this._start = cropStart;
		updatePolygon();
		invalidateAll();
	}
	
	public void setCropEnd(Point cropEnd) {
		invalidateCroppedArea();
		this._end = cropEnd;
		updatePolygon();
		invalidateAll();
	}
	
	public String getImageSource() {
		return _path;
	}
	
	public void setImageSource(String path) {
		if(_path == path) {
			return;
		}
		// update image
		invalidateAll();
		_path = path;
		refresh();
		updatePolygon();
		invalidateAll();
	}
	
	
	protected String getTagText() {
		return "@i: ";
	}
	
	private String getCrop() {
		Point start = getCropStart();
		Point end = getCropEnd();
		if(start == null || end == null) {
			return "";
		}
		return " " + start.x + " " + start.y + " " + end.x + " " + end.y;
	}
	
	// generate a fake source item to let old code work
	public Text getSource() {
		Text source = new Text(getID());
		Item.DuplicateItem(this, source);
		source.setText(this.getTagText() + this.getImageSource() + " " + this.getWidthToSave() + getCrop());
		return source;
	}
	
	public String toString() {
		return "[Image] '" + getImageSource() + "' on frame " + this.getParent().getName() + " at " + getPosition().toString();
	}

}
